// -*- C++ -*-
/*!
 * @file  pa10vel.cpp
 * @brief velocity controlled PA10 arm
 * @date $Date$
 *
 * @author FSI
 *
 * $Id$
 */

#include "pa10vel.h"

// Module specification
// <rtc-template block="module_spec">
static const char* pa10vel_spec[] =
  {
    "implementation_id", "pa10vel",
    "type_name",         "pa10vel",
    "description",       "velocity controlled PA10 arm",
    "version",           "1.0.0",
    "vendor",            "FSI",
    "category",          "pa10",
    "activity_type",     "SPORADIC",
    "kind",              "DataFlowComponent",
    "max_instance",      "10",
    "language",          "C++",
    "lang_type",         "compile",
    ""
  };
// </rtc-template>

/*!
 * @brief constructor
 * @param manager Maneger Object
 */
pa10vel::pa10vel(RTC::Manager* manager)
    // <rtc-template block="initializer">
  : RTC::DataFlowComponentBase(manager),
    m_velIn("vel", m_vel),
    m_thOut("th", m_th),
    m_PA10ProvPort("PA10Prov")

    // </rtc-template>
{
}

/*!
 * @brief destructor
 */
pa10vel::~pa10vel()
{
}

//-------------------------------------------------------------
// helper functions
//-------------------------------------------------------------
float calc_decay(float d_ftm, float l_ftm, float t_cnst) {
  if(d_ftm > l_ftm) {
    return exp((l_ftm - d_ftm)/ t_cnst);
  } else {
    return 1.0;
  }
}

/*!
 * コンポーネントを初期化する
 */
RTC::ReturnCode_t pa10vel::onInitialize()
{
  // Registration: InPort/OutPort/Service
  // <rtc-template block="registration">
  // Set InPort buffers
  addInPort("vel", m_velIn);
  
  // Set OutPort buffer
  addOutPort("th", m_thOut);
  
  // Set service provider to Ports
  m_PA10ProvPort.registerProvider("com_pa10", "ComPA10", m_com_pa10);
  
  // Set service consumers to Ports
  
  // Set CORBA Service Ports
  addPort(m_PA10ProvPort);
  
  int i;

  m_vel.data.length(7);
  m_th.data.length(7);

  for(i=0;i<7;i++) {
     ref_vel[i]=0.0;
     com_vel[i]=0.0;
  }
  long err;

  RTC_ERROR(("start pa10"));
  //待機姿勢として指定された角度へ各軸制御を行う
  err = m_com_pa10.init_pa10();
  RTC_ERROR(("ready"));
  if( err != ERR_OK ){
      return RTC::RTC_ERROR;
  }

  // </rtc-template>

  return RTC::RTC_OK;
}

/*!
 * コンポーネントを終了させる
 */

RTC::ReturnCode_t pa10vel::onFinalize()
{
  //PAライブラリの終了及びアームのクローズ処理を行う
  m_com_pa10.end_pa10();

  return RTC::RTC_OK;
}

/*
RTC::ReturnCode_t pa10vel::onStartup(RTC::UniqueId ec_id)
{
  return RTC::RTC_OK;
}
*/

/*
RTC::ReturnCode_t pa10vel::onShutdown(RTC::UniqueId ec_id)
{
  return RTC::RTC_OK;
}
*/

/*!
 * コンポーネントを活性化させる
 */

RTC::ReturnCode_t pa10vel::onActivated(RTC::UniqueId ec_id)
{
  int i;
  long err;

  //現在時刻を取得
  nTime = timeGetTime();
  oTime = nTime;

  blank_count = 0;
  d_fact = 1.0;

  for(i=0;i<7;i++) {
    ref_vel[i]=0.0;
    com_vel[i]=0.0;
  }
  //PA10の各軸速度制御の開始処理を行う
  err =  m_com_pa10.on_pa10();
  if(err != ERR_OK) {
    //PA10の各軸速度制御の一時停止処理を行う
    err =  m_com_pa10.off_pa10();
    return RTC::RTC_ERROR;
  }

  return RTC::RTC_OK;
}

/*!
 * コンポーネントを非活性化させる
 */

RTC::ReturnCode_t pa10vel::onDeactivated(RTC::UniqueId ec_id)
{
  //PA10の各軸速度制御の一時停止処理を行う
  m_com_pa10.off_pa10();

  return RTC::RTC_OK;
}

/*!
 * コンポーネントを実行する
 */

RTC::ReturnCode_t pa10vel::onExecute(RTC::UniqueId ec_id)
{
  int i;
  long err;
  ANGLE an;
  char *errstr;

  nTime = timeGetTime();
  RTC_DEBUG(("actual delta time(ms): %f",(nTime - oTime)));
  oTime = nTime;

  if( m_com_pa10.m_act_flg == 1 ){
    for(i=0;i<7;i++) {
      ref_vel[i]=0.0;
      com_vel[i]=0.0;
    }
  } else {
    if(m_velIn.isEmpty()) {
      blank_count++;
      di_tm = blank_count*(float)TIMESTEP;
      d_fact = calc_decay(di_tm, (float)0.1, (float)1.0);  // apply decay
    } else {
      blank_count = 0;
      m_velIn.read();

      for(i=0;i<7;i++) {
		//関節角度を計算しその大きさがソフトウェアリミットを超えないか確認する
        if ( fabs(m_vel.data[i]) <= m_com_pa10.m_vel_lim[i] ){
          ref_vel[i]=(float)m_vel.data[i];
        } else {
          RTC_ERROR(( "OVER LIMIT joint velocity[%d]:%f > %f [rad/sec]\n", i, m_vel.data[i] , m_com_pa10.m_vel_lim[i] ));
          ref_vel[i]=0;
        }
      }
      d_fact = 1.0;
    }
    for(i=0;i<7;i++) {
      com_vel[i]=d_fact*ref_vel[i];
    }

    err = pa_odr_vel(ARM0, com_vel);
    if(err != ERR_OK) {
		pa_err_mes(err, errstr);
		RTC_ERROR(("failed pa_odr_vel():%d %s",pa_odr_vel, errstr));
      if(err > -1000 || err <= -2000) { // not warning
		//PA10の各軸速度制御の一時停止処理を行う
        err =  m_com_pa10.off_pa10();
        return RTC::RTC_ERROR;
      }
    }

    err = pa_get_agl(ARM0, &an);

	//関節角度が正常な場合はそれをデータポート(Out)へ書き込む
    if(err == ERR_OK) {
      m_th.data[0] =  an.s1;
      m_th.data[1] =  an.s2;
      m_th.data[2] =  an.s3;
      m_th.data[3] =  an.e1;
      m_th.data[4] =  an.e2;
      m_th.data[5] =  an.w1;
      m_th.data[6] =  an.w2;
      m_thOut.write();
    }
  }

  return RTC::RTC_OK;
}

/*
RTC::ReturnCode_t pa10vel::onAborting(RTC::UniqueId ec_id)
{
  return RTC::RTC_OK;
}
*/

/*!
 * コンポーネントのエラー処理をする
 */

RTC::ReturnCode_t pa10vel::onError(RTC::UniqueId ec_id)
{
  long err;

  //PAライブラリの初期化及びアームのリセット処理を行う
  err =  m_com_pa10.reset_pa10();
  if(err != ERR_OK) {
    if(err > -1000 || err <= -2000) { // not warning
	  //PA10の各軸速度制御の一時停止処理を行う
      err =  m_com_pa10.off_pa10();
    }
  }
  return RTC::RTC_OK;
}

/*
RTC::ReturnCode_t pa10vel::onReset(RTC::UniqueId ec_id)
{
  return RTC::RTC_OK;
}
*/

/*
RTC::ReturnCode_t pa10vel::onStateUpdate(RTC::UniqueId ec_id)
{
  return RTC::RTC_OK;
}
*/

/*
RTC::ReturnCode_t pa10vel::onRateChanged(RTC::UniqueId ec_id)
{
  return RTC::RTC_OK;
}
*/



extern "C"
{
 
  void pa10velInit(RTC::Manager* manager)
  {
    coil::Properties profile(pa10vel_spec);
    manager->registerFactory(profile,
                             RTC::Create<pa10vel>,
                             RTC::Delete<pa10vel>);
  }
  
};


